"""
Modified from the pyrsistent test suite.

Pre-modification, these were MIT licensed, and are copyright:

    Copyright (c) 2022 Tobias Gustafsson

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.
"""

import pytest

from rpds import Stack


def test_literalish_works():
    assert Stack(1, 2, 3) == Stack([1, 2, 3])


def test_pop_and_peek():
    ps = Stack([1, 2])
    assert ps.peek() == 2
    assert ps.pop().peek() == 1
    assert ps.pop().pop() == Stack()


def test_instantiate_large_stack():
    assert Stack(range(1000)).peek() == 999


def test_iteration():
    assert list(Stack()) == []
    assert list(Stack([1, 2, 3]))[::-1] == [1, 2, 3]


def test_push():
    assert Stack([1, 2, 3]).push(4) == Stack([1, 2, 3, 4])


def test_push_empty_stack():
    assert Stack().push(0) == Stack([0])


def test_truthiness():
    assert Stack([1])
    assert not Stack()


def test_len():
    assert len(Stack([1, 2, 3])) == 3
    assert len(Stack()) == 0


def test_peek_illegal_on_empty_stack():
    with pytest.raises(IndexError):
        Stack().peek()


def test_pop_illegal_on_empty_stack():
    with pytest.raises(IndexError):
        Stack().pop()


def test_inequality():
    assert Stack([1, 2]) != Stack([1, 3])
    assert Stack([1, 2]) != Stack([1, 2, 3])
    assert Stack() != Stack([1, 2, 3])


def test_repr():
    assert str(Stack()) == "Stack([])"
    assert str(Stack([1, 2, 3])) in "Stack([1, 2, 3])"


def test_hashing():
    o = object()

    assert hash(Stack([o, o])) == hash(Stack([o, o]))
    assert hash(Stack([o])) == hash(Stack([o]))
    assert hash(Stack()) == hash(Stack([]))
    assert not (hash(Stack([1, 2])) == hash(Stack([1, 3])))
    assert not (hash(Stack([1, 2])) == hash(Stack([2, 1])))
    assert not (hash(Stack([o])) == hash(Stack([o, o])))
    assert not (hash(Stack([])) == hash(Stack([o])))

    assert hash(Stack([1, 2])) != hash(Stack([1, 3]))
    assert hash(Stack([1, 2])) != hash(Stack([2, 1]))
    assert hash(Stack([o])) != hash(Stack([o, o]))
    assert hash(Stack([])) != hash(Stack([o]))
    assert not (hash(Stack([o, o])) != hash(Stack([o, o])))
    assert not (hash(Stack([o])) != hash(Stack([o])))
    assert not (hash(Stack([])) != hash(Stack([])))


def test_sequence():
    m = Stack("asdf")
    assert m == Stack(["a", "s", "d", "f"])


# Non-pyrsistent-test-suite tests


def test_more_eq():
    o = object()

    assert Stack([o, o]) == Stack([o, o])
    assert Stack([o]) == Stack([o])
    assert Stack() == Stack([])
    assert not (Stack([1, 2]) == Stack([1, 3]))
    assert not (Stack([o]) == Stack([o, o]))
    assert not (Stack([]) == Stack([o]))

    assert Stack([1, 2]) != Stack([1, 3])
    assert Stack([o]) != Stack([o, o])
    assert Stack([]) != Stack([o])
    assert not (Stack([o, o]) != Stack([o, o]))
    assert not (Stack([o]) != Stack([o]))
    assert not (Stack() != Stack([]))


def test_rpds_doc():
    """
    From the rpds docs.
    """
    stack = Stack().push("stack")
    assert stack.peek() == "stack"

    a_stack = stack.push("a")
    assert a_stack.peek() == "a"

    stack_popped = a_stack.pop()
    assert stack_popped == stack



from pathlib import Path
from tempfile import TemporaryDirectory
import os

import nox

ROOT = Path(__file__).parent
PYPROJECT = ROOT / "pyproject.toml"
DOCS = ROOT / "docs"
TESTS = ROOT / "tests"

SUPPORTED = [
    "3.11",
    "pypy3.11",
    "3.12",
    "3.13",
    "3.14t",
    "3.14",
    "3.15t",
    "3.15",
]
LATEST = SUPPORTED[-1]

nox.options.default_venv_backend = "uv"
nox.options.sessions = []


def session(default=True, python=LATEST, **kwargs):  # noqa: D103
    def _session(fn):
        if default:
            nox.options.sessions.append(kwargs.get("name", fn.__name__))
        return nox.session(python=python, **kwargs)(fn)

    return _session


@session(python=SUPPORTED)
def tests(session):
    """
    Run the test suite with a corresponding Python version.
    """
    # Really we want --profile=test here (for
    # https://github.com/crate-py/rpds/pull/87#issuecomment-2291409297)
    # but it produces strange symbol errors saying:
    #   dynamic module does not define module export function (PyInit_rpds)
    # so OK, dev it is.
    session.run_install(
        "uv",
        "sync",
        "--group=test",
        "--config-setting",
        "build-args=--profile=dev",
        "--no-cache",
        f"--python={session.virtualenv.location}",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )

    if session.posargs and session.posargs[0] == "coverage":
        if len(session.posargs) > 1 and session.posargs[1] == "github":
            github = Path(os.environ["GITHUB_STEP_SUMMARY"])
        else:
            github = None

        session.install("coverage[toml]")
        session.run("coverage", "run", "-m", "pytest", TESTS)
        if github is None:
            session.run("coverage", "report")
        else:
            with github.open("a") as summary:
                summary.write("### Coverage\n\n")
                summary.flush()  # without a flush, output seems out of order.
                session.run(
                    "coverage",
                    "report",
                    "--format=markdown",
                    stdout=summary,
                )
    else:
        session.run("pytest", "--parallel-threads=10", *session.posargs, TESTS)


@session(tags=["build"])
def build(session):
    """
    Build a distribution suitable for PyPI and check its validity.
    """
    session.install("build[uv]", "twine")
    with TemporaryDirectory() as tmpdir:
        session.run(
            "pyproject-build",
            "--installer=uv",
            ROOT,
            "--outdir",
            tmpdir,
        )
        session.run("twine", "check", "--strict", tmpdir + "/*")


@session(tags=["style"])
def style(session):
    """
    Check Python code style.
    """
    session.install("ruff")
    session.run("ruff", "check", TESTS, __file__)


@session()
def typing(session):
    """
    Check the codebase using pyright by type checking the test suite.
    """
    session.run_install(
        "uv",
        "sync",
        "--group=typing",
        "--config-setting",
        "build-args=--profile=dev",
        "--no-cache",
        f"--python={session.virtualenv.location}",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )
    session.run("pyright", TESTS)


@session(tags=["docs"])
@nox.parametrize(
    "builder",
    [
        nox.param(name, id=name)
        for name in [
            "dirhtml",
            "doctest",
            "linkcheck",
            "man",
            "spelling",
        ]
    ],
)
def docs(session, builder):
    """
    Build the documentation using a specific Sphinx builder.
    """
    session.run_install(
        "uv",
        "sync",
        "--group=docs",
        "--config-setting",
        "build-args=--profile=dev",
        "--no-cache",
        f"--python={session.virtualenv.location}",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )
    with TemporaryDirectory() as tmpdir_str:
        tmpdir = Path(tmpdir_str)
        argv = ["-n", "-T", "-W"]
        if builder != "spelling":
            argv += ["-q"]
        posargs = session.posargs or [tmpdir / builder]
        session.run(
            "python",
            "-m",
            "sphinx",
            "-b",
            builder,
            DOCS,
            *argv,
            *posargs,
        )


@session(tags=["docs", "style"], name="docs(style)")
def docs_style(session):
    """
    Check the documentation style.
    """
    session.install(
        "doc8",
        "pygments",
        "pygments-github-lexers",
    )
    session.run("python", "-m", "doc8", "--config", PYPROJECT, DOCS)
